use std::sync::Arc;

use tracing::debug;

use ubisync_lib::{
    api::events::AppEvent,
    peer::Peer,
    types::{Element, ElementContent, ElementId, ContentUpdateStrategy, MessageId, PotId},
};

use crate::node_events::UbisyncNodeEvent;

use super::{database::StateDB, State};

pub struct CommState {
    state: Arc<State>,
}

impl CommState {
    pub fn new(state: Arc<State>) -> Self {
        CommState { state: state }
    }

    pub fn add_received_element(
        &self,
        id: ElementId,
        content: ElementContent,
        update_strategy: ContentUpdateStrategy,
        latest_message: Option<MessageId>,
        pot_id: PotId,
    ) -> anyhow::Result<()> {
        self.db()
            .add_element(
                id.clone(),
                content,
                update_strategy,
                latest_message,
                false,
                pot_id,
            )
            .inspect(|_| debug!("Added element {{{}}}", id.to_string()))
    }

    pub fn update_element_content(
        &self,
        id: ElementId,
        content: ElementContent,
        latest_message: MessageId,
    ) -> anyhow::Result<()> {
        //TODO: resolve potential conflicts with local changes
        self.db().set_element_content(id.clone(), content)?;
        self.db()
            .set_element_latest_message(id.clone(), Some(latest_message))?;
        debug!("Updated element {{{}}}", id.to_string());

        if let Some(el) = self.db().get_element(id.clone()) {
            if let Some(pot) = el.pot() {
                if let Ok(apps) = self.db().get_pot_members(pot.to_owned()) {
                    for app in apps {
                        self.state
                            .emit_app_event(&app, AppEvent::ElementUpdate { id: id.clone() })
                    }
                }
            }
        }

        Ok(())
    }

    pub fn remove_element(&self, id: ElementId) -> anyhow::Result<()> {
        self.db()
            .remove_element(id.clone())
            .inspect(|_| debug!("Removed element {{{}}}", &id.to_string()))
    }

    pub fn get_element(&self, id: ElementId) -> Option<Element> {
        self.db().get_element(id)
    }

    pub fn set_peer(&self, peer: Peer) -> anyhow::Result<()> {
        self.db()
            .add_peer(peer.clone())
            .inspect(|_| debug!("Added peer {:?}.", peer))
    }

    pub fn get_peers(&self) -> anyhow::Result<Vec<Peer>> {
        self.state.get_peers()
    }

    pub fn add_pot(&self, id: PotId, app_type: String) -> anyhow::Result<()> {
        self.db().add_pot(id.clone(), app_type.clone())?;

        let _ = self.state.emit_node_event(UbisyncNodeEvent::NewPot {
            id: id,
            app_type: app_type,
        });

        Ok(())
    }

    fn db(&self) -> &StateDB {
        &self.state.db
    }
}

#[cfg(test)]
mod tests {
    use crate::state::State;

    use super::CommState;

    use tracing::Level;
    use ubisync_lib::types::{ElementContent, ElementId, ContentUpdateStrategy, MessageId, PotId};

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_add() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = CommState::new(State::new(None).await.unwrap());
        let id = ElementId::new();
        let pot_id = PotId::new();
        state
            .add_received_element(
                id.clone(),
                ElementContent::Text("Test-text".to_string()),
                ContentUpdateStrategy::Overwrite,
                Some(MessageId::new()),
                pot_id,
            )
            .unwrap();
        let el = state.get_element(id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text".to_string()),
            el.content().to_owned()
        )
    }

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_update() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = CommState::new(State::new(None).await.unwrap());
        let id = ElementId::new();
        let pot_id = PotId::new();
        state
            .add_received_element(
                id.clone(),
                ElementContent::Text("Test-text".to_string()),
                ContentUpdateStrategy::Overwrite,
                Some(MessageId::new()),
                pot_id,
            )
            .unwrap();
        state
            .update_element_content(
                id.clone(),
                ElementContent::Text("Test-text 2".to_string()),
                MessageId::new(),
            )
            .unwrap();
        let el = state.get_element(id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text 2".to_string()),
            el.content().to_owned()
        )
    }
}
